xquery version "3.1";
(:~
 : This module creates a GraphML representation of one or more given modules or
 : even the whole SADE app. To achieve this we take the information retrieved by
 : inspect:inspect-module and convert the relevant nodes to GraphML.
 :
 : The process was roughly inspired by the graphviz app for eXist-db which can
 : be found at https://github.com/KitWallace/graphviz.
 :
 : @author Michelle Weidling
 : @since v3.1.1
 : @see https://github.com/KitWallace/graphviz
 : @see http://graphml.graphdrawing.org/specification.html
 :
 : :)


module namespace inspect2graphml="https://sade.textgrid.de/ns/inspect2graphml";

import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace inspect="http://exist-db.org/xquery/inspection";

declare namespace graphml="http://graphml.graphdrawing.org/xmlns";
declare namespace xsi="http://www.w3.org/2001/XMLSchema-instance";


(:~
 : The interface for other modules.
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @param $option Selects the modules which should be displayed in the output SVG file (cf. callgraph:main for valid options)
 : @return The graph representation as element(graphml:graphml)
 : @return The path to the stored GraphML file
 :  :)
declare function inspect2graphml:main($option as xs:string, $filename
as xs:string) as item()+ {
    let $graphml := inspect2graphml:get-graphml($option)
    return
        ($graphml,
        xmldb:store($config:app-root || "modules/callgraph/graphml/", $filename || "-graphml.xml", $graphml))
};


(:~
 : Creates a GraphML representation of one or more given modules or even the
 : whole SADE app.
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @param $option Selects the modules which should be displayed in the output SVG file (cf. callgraph:main for valid options)
 :  :)
declare function inspect2graphml:get-graphml($option as xs:string) as
element(graphml:graphml) {
    let $xqms := inspect2graphml:get-xqms($option)
    return
        if(count($xqms) = 1) then
            element graphml:graphml {
                attribute xsi:schemaLocation {"http://graphml.graphdrawing.org/xmlns http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd"},
                let $inspect-result := inspect:inspect-module(xs:anyURI($xqms))
                let $transformed-subgraph := inspect2graphml:transform($inspect-result, true())
                return local:sort-edges-and-nodes($transformed-subgraph)
            }

        else
            element graphml:graphml {
                attribute xsi:schemaLocation {"http://graphml.graphdrawing.org/xmlns http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd"},
                element graphml:graph {
                    attribute id {"base-graph"},
                    attribute edgedefault {"directed"},
                    for $xqm in $xqms return
                        let $inspect-result := inspect:inspect-module(xs:anyURI($xqm))
                        let $transformed-subgraph := inspect2graphml:transform($inspect-result, false())

                        return
                            element graphml:node {
                                attribute id {util:hash($xqm, "md5")},
                                local:sort-edges-and-nodes($transformed-subgraph)
                            }
                }
            }
};


(:~
 : Transforms the relevant nodes that have been generated by inspect:inspect-module
 : to GraphML elements.
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @param $nodes The nodes that are the result of inspect:inspect-module
 : @param $single-module A flag denoting if the user wants only a single module to be considered. In this case all function elements have to be converted to graphml:node which otherwise would lead to redundancies.
 : @return The transformed node(s)
 :  :)
declare function inspect2graphml:transform($nodes as node()*, $single-module as xs:boolean) as node()* {
    for $node in $nodes return
        typeswitch ($node)

        case element(module) return
                element graphml:graph {
                    attribute id {$node/@prefix},
                    attribute edgedefault {"directed"},
                    inspect2graphml:transform($node/node(), $single-module)
                }

        case element(function) return
            (local:make-node($node),
            inspect2graphml:transform($node/node(), $single-module))

        case element(calls) return
            let $source-function := $node/ancestor::function[1]
            let $prefix := substring-before($source-function/@name, ":")
            for $called in $node/function
                return
                    (element graphml:edge {
                        attribute source {$source-function/@name},
                        attribute target {$called/@name}
                    },
                    if($single-module
                    and not(substring-before($called/@name, ":") = $prefix)) then
                        local:make-node($called)
                    else
                        ())

        default return
            ()
};


(:~
 : Gets all relevant *.xqm files.
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @param $option Selects the modules which should be displayed in the output SVG file (cf. callgraph:main for valid options)
 : @return A sequence of module URIs
 :  :)
declare function inspect2graphml:get-xqms($option as xs:string) as xs:string* {
    switch ($option)
        case "full" return (collection($config:app-root)/base-uri())[ends-with(., ".xqm")]

        default return
            if(ends-with($option, ".xqm")) then
                $option
            else
                for $child in xmldb:get-child-resources($option) return
                    if(contains($child, ".xqm")) then
                        $option || "/" || $child
                    else
                        ()
};


(:~
 : Creates a graphml:node element.
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @param $node The current result node of inspect:inspect-module
 : @return A respective graphml:node element
 :  :)
declare function local:make-node($node as node()) as element(graphml:node) {
    element graphml:node {
        attribute id {$node/@name}
    }
};


(:~
 : As the output of inspect2graphml:transform mixes up graphml:node and
 : graphml:edge, this function sorts them: We want all graphml:node to be listed
 : first, followed by all graphml:edge elements. The order depend on the node ID
 : and the source and target attribute respectively.
 :
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @param $graph A (sub-)graph with nodes and edges
 : @return A graph with sorted nodes and edges
 :  :)
declare function local:sort-edges-and-nodes($graph as element(graphml:graph))
as element(graphml:graph) {
    element graphml:graph {
        $graph/@*,
        for $node in $graph/graphml:node
            order by $node/@id/string() ascending
            return $node,
        for $edge in $graph/graphml:edge
            order by $edge/@source, $edge/@target
            return $edge
    }
};
